# -*- coding: utf-8 -*-

# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import ast
from ast import Pass, Module, Bytes, copy_location, Call, Name, Load, Str, BitOr
from ast import fix_missing_locations, Import, alias, Attribute, ImportFrom
from ast import keyword, Gt, Lt, GtE, RShift, Tuple
import pickle
from collections import defaultdict
import ayrton

def append_to_tuple (t, o):
    l= list (t)
    l.append (o)
    return tuple (l)

def pop_from_tuple (t, n=-1):
    l= list (t)
    l.pop (n)
    return tuple (l)

def is_executable (node):
    # Call(func=Call(func=Name(id='Command', ctx=Load()),
    #                args=[Str(s='ls')], keywords=[], starargs=None, kwargs=None),
    #      args=[Str(s='-l')], keywords=[], starargs=None, kwargs=None)
    return (type (node)==Call and
            type (node.func)==Call and
            type (node.func.func)==Name and
            node.func.func.id=='Command')

def has_keyword (node, keyword):
    return any ([kw.arg==keyword for kw in node.keywords])

class CrazyASTTransformer (ast.NodeTransformer):
    def __init__ (self, environ):
        super ().__init__ ()
        self.environ= environ
        self.known_names= defaultdict (lambda: 0)
        self.stack= ()
        self.defined_names= defaultdict (lambda: [])

    # The following constructs bind names:
    # [x] formal parameters to functions,
    # [x] import statements,
    # [x] class and
    # [x] function definitions (these bind the class or function name
    #     in the defining block),
    # [x] and targets that are identifiers if occurring in an assignment,
    # [x] for loop header, or
    # [x] after as in a with statement
    # [ ] or except clause.
    # The import statement of the form from ... import * binds all names defined
    # in the imported module, except those beginning with an underscore.

    # A block is a piece of Python program text that is executed as a unit.
    # The following are blocks:
    # [ ] a module,
    # [x] a function body, and
    # [x] a class definition.
    # [ ] A script file is a code block.
    # [ ] The string argument passed to the built-in functions eval() and exec()
    #     is a code block.

    # A scope defines the visibility of a name within a block. If a local variable
    # is defined in a block, its scope includes that block. If the definition
    # occurs in a function block, the scope extends to any blocks contained within
    # the defining one, unless a contained block introduces a different binding
    # for the name. The scope of names defined in a class block is limited to the
    # class block; it does not extend to the code blocks of methods – this
    # includes comprehensions and generator expressions since they are implemented
    # using a function scope.

    # [x] A target occurring in a del statement is also considered bound for this
    #     purpose (though the actual semantics are to unbind the name).

    def visit_Import (self, node):
        self.generic_visit (node)
        # Import(names=[alias(name='foo', asname=None)])
        for name in node.names:
            if name.asname is not None:
                n= name.asname
            else:
                n= name.name
            self.known_names[n]+=1
            self.defined_names[self.stack].append (n)

        return node

    visit_ImportFrom= visit_Import
        # ImportFrom(module='bar', names=[alias(name='baz', asname=None)], level=0)

    def visit_ClassDef (self, node):
        # ClassDef(name='foo', bases=[], keywords=[], starargs=None, kwargs=None,
        #          body=[Pass()], decorator_list=[])
        self.stack= append_to_tuple (self.stack, node.name)
        self.generic_visit (node)

        # take out the function from the stack
        names= self.defined_names[self.stack]
        self.stack= pop_from_tuple (self.stack)
        # ... and remove the names defined in it from the known_names
        for name in names:
            self.known_names[name]-= 1

        return node

    def visit_FunctionDef (self, node):
        # FunctionDef(name='foo',
        #             args=arguments(args=[arg(arg='a', annotation=None),
        #                                  arg(arg='b', annotation=None)],
        #                            vararg='args', varargannotation=None,
        #                            kwonlyargs=[], kwarg='kwargs',
        #                            kwargannotation=None, defaults=[Num(n=1)],
        #                            kw_defaults=[]),
        #             body=[Pass()], decorator_list=[], returns=None)
        self.stack= append_to_tuple (self.stack, node.name)

        # add the arguments as local names
        for arg in node.args.args:
            self.known_names[arg.arg]+= 1
            self.defined_names[self.stack].append (arg.arg)

        self.generic_visit (node)

        # take out the function from the stack
        names= self.defined_names[self.stack]
        self.stack= pop_from_tuple (self.stack)
        # ... and remove the names defined in it from the known_names
        for name in names:
            self.known_names[name]-= 1

        return node

    def visit_For (self, node):
        # For(target=Name(id='x', ctx=Store()), iter=List(elts=[], ctx=Load()),
        #     body=[Pass()], orelse=[])
        self.known_names[node.target.id]+= 1
        self.defined_names[self.stack].append (node.target.id)

        self.generic_visit (node)

        return node

    def visit_ExceptHandler (self, node):
        # Try(body=[Expr(value=Name(id='fo', ctx=Load()))],
        #     handlers=[ExceptHandler(type=Name(id='A', ctx=Load()), name='e',
        #               body=[Pass()])],
        #     orelse=[], finalbody=[])

        # not sure what/how to do here
        # self.known_names[name.id]+= 1
        # self.defined_names[self.stack].append (name.id)
        self.generic_visit (node)

        return node

    def assign (self, node):
        if type (node)==Name:
            # Assign(targets=[Name(id='a', ctx=Store())], value=Num(n=2))
            self.known_names[node.id]+= 1
            self.defined_names[self.stack].append (node.id)
        elif type (node)==Tuple:
            # Assign(targets=[Tuple(elts=[Name(id='a', ctx=Store()), Name(id='b', ctx=Store())], ...
            for elt in node.elts:
                self.assign (elt)

    def visit_Assign (self, node):
        self.generic_visit (node)
        for target in node.targets:
            self.assign (target)

        return node

    def visit_Delete (self, node):
        self.generic_visit (node)
        # Delete(targets=[Name(id='foo', ctx=Del())])
        for target in node.targets:
            self.known_names[target.id]-= 1
            self.defined_names[self.stack].remove (target.id)

        return node

    # BinOp(left=BinOp(left=Name(id='a', ctx=Load()),
    #                  op=BitOr(), right=Name(id='b', ctx=Load())),
    #       op=BitOr(), right=Name(id='c', ctx=Load()))
    def visit_BinOp (self, node):
        self.generic_visit (node)

        # BinOp( left=Call(...), op=BitOr(), right=Call(...))
        if type (node.op)==BitOr:
            # pipe
            # BinOp (left, BitOr, right) -> right (left, ...)

            # check the left and right; if they're calls to Command
            # then do the magic

            # NOTE: they're Commands only because the children have already been
            # visited and transformed to such things

            both= is_executable (node.left) and is_executable (node.right)

            if both:
                # left (...) | right (...)

                # r, w= os.pipe ()
                # left (..., _out=w)
                # os.close (w)
                # right (..., _in=r)
                # os.close (r)

                # Assign(targets=[Tuple(elts=[Name(id='r', ctx=Store()),
                #                             Name(id='w', ctx=Store())], ctx=Store())],
                #        value=Call(func=Attribute(value=Name(id='os', ctx=Load()),
                #                                  attr='pipe', ctx=Load()),
                #                   args=[], keywords=[], starargs=None, kwargs=None)),
                # Expr(value=Call(func=Name(id='echo', ctx=Load()), args=[Str(s='pipe!')], keywords=[keyword(arg='_out', value=Name(id='w', ctx=Load()))], starargs=None, kwargs=None)), Expr(value=Call(func=Attribute(value=Name(id='os', ctx=Load()), attr='close', ctx=Load()), args=[Name(id='w', ctx=Load())], keywords=[], starargs=None, kwargs=None)), Expr(value=Call(func=Name(id='grep', ctx=Load()), args=[Str(s='pipe')], keywords=[keyword(arg='_in', value=Name(id='r', ctx=Load()))], starargs=None, kwargs=None)), Expr(value=Call(func=Attribute(value=Name(id='os', ctx=Load()), attr='close', ctx=Load()), args=[Name(id='r', ctx=Load())], keywords=[], starargs=None, kwargs=None))

                # I can't believe it's this easy
                # TODO: check if _err is not being captured instead
                node.left.keywords.append (keyword (arg='_out',
                                                    value=Name (id='Capture', ctx=Load ())))
                ast.fix_missing_locations (node.left)
                node.right.keywords.append (keyword (arg='_in', value=node.left))
                node= node.right

                # Call(func=Call(func=Attribute(value=Name(id='CommandWrapper', ctx=Load()),
                #                               attr='_create', ctx=Load()),
                #                args=[Str(s='grep')], keywords=[], starargs=None, kwargs=None),
                #      args=[Call(func=Call(func=Attribute(value=Name(id='CommandWrapper', ctx=Load()),
                #                                          attr='_create', ctx=Load()),
                #                      args=[Str(s='ls')], keywords=[], starargs=None, kwargs=None),
                #                 args=[], keywords=[keyword(arg='_out',
                #                                            value=Name(id='Capture', ctx=Load()))],
                #                 starargs=None, kwargs=None),
                #            Str(s='foo')],
                #      keywords=[], starargs=None, kwargs=None)

        elif type (node.op)==RShift:
            # BinOp(left=Call(func=Name(id='ls', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=None),
            #       op=RShift(),
            #       right=Str(s='foo.txt'))
            if is_executable (node.left):
                node.left.keywords.append (keyword (arg='_out',
                                                    value=Call (func=Name (id='open', ctx=Load ()),
                                                                args=[node.right, Str (s='ab')], keywords=[], starargs=None, kwargs=None)))
                ast.fix_missing_locations (node.left)
                node= node.left

        return node

    # Compare(left=Call(func=Name(id='grep', ctx=Load()), args=[Str(s='root')], keywords=[], starargs=None, kwargs=None),
    #        ops=[Lt()],
    #        comparators=[BinOp(left=Str(s='/etc/passwd'),
    #                           op=RShift(),
    #                           right=Str(s='/tmp/foo'))])
    def visit_Compare (self, node):
        self.generic_visit (node)
        ans= node

        # ls () < "bar.txt" > "foo.txt"
        # Compare(left=Call(func=Name(id='ls', ctx=Load()), args=[], keywords=[],
        #                        starargs=None, kwargs=None),
        #         ops=[Lt(), Gt()],
        #         comparators=[Str(s='bar.txt'), Str(s='foo.txt')]))

        if is_executable (node.left):
            # yes, they're reversed, but it makes more sense to me like this
            for comp, op in zip (node.ops, node.comparators):
                if type (comp)==Gt:
                    # > means _out
                    node.left.keywords.append (keyword (arg='_out', value=op))
                    ans= node.left

                if type (comp)==GtE:
                    # >= means _out+_err
                    node.left.keywords.append (keyword (arg='_out', value=op))
                    node.left.keywords.append (keyword (arg='_err_to_out', value=op))
                    ans= node.left

                elif type (comp)==Lt:
                    # < means _in

                    # now, _in works differently
                    # a string is written directly to the stdin,
                    # instead of creating a file() with that name
                    # so, we do it ourseleves.
                    if type (op)==Str:
                        op= Call (func=Name (id='open', ctx=Load ()), args=[op],
                                  keywords=[], starargs=None, kwargs=None)

                    node.left.keywords.append (keyword (arg='_in', value=op))
                    ast.fix_missing_locations (node.left)
                    ans= node.left

        return ans

    def visit_Call (self, node):
        self.generic_visit (node)
        # Call(func=Name(id='b', ctx=Load()), args=[], keywords=[], starargs=None,
        #      kwargs=None)
        if type (node.func)==ast.Name:
            func_name= node.func.id
            defs= self.known_names[func_name]
            if defs==0:
                try:
                    # first check if it's not one of the builtin functions
                    self.environ[func_name]
                except KeyError:
                    # I guess I have no other option but to try to execute
                    # something here...
                    new_node= Call (func=Name (id='Command', ctx=Load ()),
                                    args=[Str (s=func_name)], keywords=[],
                                    starargs=None, kwargs=None)

                    # check if the first parameter is a Command; if so, redirect
                    # its output, remove it from the args and put it in the _in
                    # kwarg
                    if len (node.args)>0:
                        first_arg= node.args[0]
                        if is_executable (first_arg) and not has_keyword (first_arg, '_err'):
                            first_arg.keywords.append (
                                keyword (arg='_out', value=Name (id='Capture', ctx=Load ()))
                                )
                            node.args.pop (0)
                            node.keywords.append (keyword (arg='_in', value=first_arg))

                    ast.copy_location (new_node, node)
                    node.func= new_node
                    ast.fix_missing_locations (node)

        return node

    def visit_With (self, node):
        # With(items=[withitem(context_expr=Call(func=Name(id='foo', ctx=Load()),
        #                                        args=[], keywords=[], starargs=None,
        #                                        kwargs=None),
        #                      optional_vars=Name(id='bar', ctx=Store()))], body=[Pass()])
        for item in node.items:
            if item.optional_vars is not None:
                self.known_names[item.optional_vars.id]+= 1
                self.defined_names[self.stack].append (item.optional_vars.id)

        self.generic_visit (node)

        # handle 'remote'
        sub_node= node.items[0].context_expr
        if (type (sub_node)==Call and hasattr (sub_node.func, 'id') and
            sub_node.func.id=='remote'):
            # capture the body and put it as the first argument to ssh()
            # but within a module, and already pickled;
            # otherwise we need to create an AST for the call of all the
            # constructors in the body we capture... it's complicated
            m= Module (body=node.body)

            data= pickle.dumps (m)
            s= Bytes (s=data)
            s.lineno= node.lineno
            s.col_offset= node.col_offset
            sub_node.args.insert (0, s)

            p= Pass ()
            p.lineno= node.lineno+1
            p.col_offset= node.col_offset+4

            node.body= [ p ]

        return node



# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import io
from collections.abc import Iterable
import ayrton

encoding= sys.getdefaultencoding ()

# special value to signal that the output should be captured
# instead of going to stdout
Capture= (42, )

class o (object):
    def __init__ (self, **kwargs):
        option= list (kwargs.items ())[0]
        self.key=   option[0]
        self.value= option[1]

class CommandFailed (Exception):
    def __init__ (self, code):
        self.code= code

class CommandNotFound (Exception):
    def __init__ (self, path):
        self.path= path

class Command:
    default_options= dict (
        _in_tty= False,
        _out_tty= False,
        _env= {},
        _end= os.linesep.encode (encoding),
        _chomp= True,
        _encoding= encoding,
        )

    supported_options= ('_in', '_out', '_err', '_end', '_chomp', '_encoding',
                        '_env',)

    def __init__ (self, path):
        self.path= path

        self.stdin_pipe= None
        self.stdout_pipe= None
        self.stderr_pipe= None

        self.exit_code= None
        self.capture_file= None

    def prepare_fds (self):
        if '_in' in self.options:
            i= self.options['_in']
            if not isinstance (i, io.IOBase) and type (i)!=int and i is not None:
                if self.options['_in_tty']:
                    # TODO: no support yet for input from file when _in_tty
                    # NOTE: os.openpty() returns (master, slave)
                    # but when writing master=w, slave=r
                    (master, slave)= os.openpty ()
                    self.stdin_pipe= (slave, master)
                else:
                    self.stdin_pipe= os.pipe ()

        if '_out' in self.options and self.options['_out']==Capture:
            if self.options['_out_tty']:
                if self.options['_in_tty']:
                    # we use a copy of the pty created for the stdin
                    self.stdout_pipe= (os.dup (self.stdin_pipe[1]),
                                       os.dup (self.stdin_pipe[0]))
                else:
                    # this time the order is right
                    (master, slave)= os.openpty ()
                    self.stdout_pipe= (master, slave)
            else:
                self.stdout_pipe= os.pipe ()

        if '_err' in self.options and self.options['_err']==Capture:
            # if stdout is also Capture'd, then use the same pipe
            if not '_out' in self.options or self.options['_out']!=Capture:
                # if stdout is a tty, hook to that one
                if self.options['_out_tty']:
                    self.stderr_pipe= (os.dup (self.stdout_pipe[0]),
                                       os.dup (self.stdout_pipe[1]))
                else:
                    self.stderr_pipe= os.pipe ()

    def child (self, cmd, *args, **kwargs):
        if '_in' in self.options:
            i= self.options['_in']
            if i is None:
                # connect to /dev/null
                # it's not /dev/zero, see man (4) zero
                i= open (os.devnull, 'rb')

            if isinstance (i, io.IOBase):
                # this does not work with file like objects
                # dup its fd int stdin (0)
                os.dup2 (i.fileno (), 0)
            elif type (i)==int:
                os.dup2 (i, 0)
            else:
                r, w= self.stdin_pipe
                os.dup2 (r, 0)
                os.close (w)

        if '_out' in self.options:
            o= self.options['_out']
            if o is None:
                # connect to /dev/null
                o= open (os.devnull, 'wb') # no need to create it

            if isinstance (o, io.IOBase):
                # this does not work with file like objects
                # dup its fd in stdout (1)
                os.dup2 (o.fileno (), 1)
            elif type (o)==int:
                os.dup2 (o, 1)
            elif o==Capture:
                r, w= self.stdout_pipe
                os.dup2 (w, 1)
                os.close (r)
            elif type (o) in (bytes, str):
                f= open (o, 'w+')
                os.dup2 (f.fileno (), 1)
                f.close ()

        if '_err' in self.options:
            e= self.options['_err']
            if e is None:
                # connect to /dev/null
                e= open (os.devnull, 'wb') # no need to create it

            if isinstance (e, io.IOBase):
                # this does not work with file like objects
                # dup its fd int stderr (2)
                os.dup2 (e.fileno (), 2)

            if e==Capture:
                if '_out' in self.options and self.options['_out']==Capture:
                    # send it to the same pipe as stdout
                    r, w= self.stdout_pipe
                    os.dup2 (w, 2)
                else:
                    r, w= self.stderr_pipe
                    os.dup2 (w, 2)
                    os.close (r)

        # args is a tuple
        args= self.prepare_args (cmd, args, kwargs)
        try:
            os.execvpe (cmd, args, self.options['_env'])
        except FileNotFoundError:
            os._exit (127)

    def prepare_args (self, cmd, args, kwargs):
        ans= [cmd]

        for arg in args:
            if type (arg)==o:
                self.prepare_arg (ans, arg.key, arg.value)
            else:
                ans.append (arg)

        for k, v in kwargs.items ():
            self.prepare_arg (ans, k, v)

        return ans

    def prepare_arg (self, seq, name, value):
        if value!=False:
            if len (name)==1:
                arg="-%s" % name
            else:
                # TODO: longopt_prefix
                # and/or simply subclass find(Command)
                arg="--%s" % name
            seq.append (arg)

            if value!=True:
                seq.append (str (value))

    def parent (self, cmd, child_pid):
        reader_pipe= None

        if self.stdin_pipe is not None:
            # str -> write into the fd
            # list -> write each
            # file -> take fd
            i= self.options['_in']

            r, w= self.stdin_pipe
            os.close (r)

            if type (i)==str:
                os.write (w, i.encode (encoding))
                os.write (w, self.options['_end'])
            elif type (i)==bytes:
                os.write (w, i)
                os.write (w, self.options['_end'])
            elif isinstance (i, Iterable):
                for e in i:
                    os.write (w, str (e).encode (encoding))
                    os.write (w, self.options['_end'])
            else:
                os.write (w, str (i).encode (encoding))
                os.write (w, self.options['_end'])

            os.close (w)

        self.exit_code= os.waitpid (child_pid, 0)[1] >> 8

        if self.stdout_pipe is not None:
            # this will also read stderr if both are Capture'd
            reader_pipe= self.stdout_pipe
        if self.stderr_pipe is not None:
            reader_pipe= self.stderr_pipe

        if reader_pipe is not None:
            r, w= reader_pipe
            os.close (w)
            self.capture_file= open (r)

        if self.exit_code==127:
            raise CommandNotFound (cmd)

        if ayrton.runner.options.get ('errexit', False) and self.exit_code!=0:
            raise CommandFailed (self)

    def __call__ (self, *args, **kwargs):
        self.options= self.default_options.copy ()
        for option in self.supported_options:
            try:
                # update with the passed value
                self.options[option]= kwargs[option]
                # we don't need the option anymore
                del kwargs[option]
            except KeyError:
                # ignore
                pass

        self.options['_env'].update (ayrton.runner.environ.os_environ)

        self.stdin_pipe= None
        self.stdout_pipe= None
        self.stderr_pipe= None

        self.exit_code= None
        self.capture_file= None

        self.prepare_fds ()

        if type (self.options['_end'])!=bytes:
            self.options['_end']= str (self.options['_end']).encode (encoding)

        r= os.fork ()
        if r==0:
            self.child (self.path, *args, **kwargs)
        else:
            self.parent (self.path, r)

        return self

    def __bool__ (self):
        return self.exit_code==0

    def __str__ (self):
        return self.capture_file.read ()

    def __iter__ (self):
        if self.capture_file is not None:
            for line in self.capture_file.readlines ():
                if self.options['_chomp']:
                    line= line.rstrip (os.linesep)

                yield line

            # finish him!
            self.capture_file.close ()
        else:
            # TODO
            pass

    # BUG this method is leaking an opend file()
    # self.capture_file
    def readline (self):
        line= self.capture_file.readline ()
        if self.options['_chomp']:
            line= line.rstrip (os.linesep)

        return line

    def close (self):
        self.capture_file.close ()

    def readlines (self):
        # ugly way to not leak the file()
        return ( line for line in self )



# -*- coding: utf-8 -*-

# (c) 2013 Marcos Dione <mdione@grulic.org.ar>

# This file is part of ayrton.
#
# ayrton is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ayrton is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ayrton.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import importlib
import builtins
import pickle
import ast
from ast import fix_missing_locations, alias, ImportFrom
import traceback

# things that have to be defined before importing ayton.execute :(
# singleton
runner= None

from ayrton.castt import CrazyASTTransformer
from ayrton.execute import o, Command, Capture, CommandFailed

__version__= '0.4'

class Environment (object):
    # this class handles not only environment variables
    # but also locals, globals and python and ayrton builtins
    # the latter are only other python functions that are 'promoted' to builtins
    def __init__ (self, globals=None, locals=None, **kwargs):
        super ().__init__ ()

        if globals is None:
            self.globals= {}
        else:
            self.globals= globals

        if locals is None:
            self.locals= {}
        else:
            self.locals= locals

        self.python_builtins= builtins.__dict__.copy ()
        self.ayrton_builtins= {}
        polute (self.ayrton_builtins)
        self.os_environ= os.environ.copy ()

        # now polute the locals with kwargs
        for k, v in kwargs.items ():
            # BUG: this sucks
            if k=='argv':
                self.ayrton_builtins['argv']= v
            else:
                self.locals[k]= v

    def __getitem__ (self, k):
        strikes= 0
        for d in (self.locals, self.globals, self.os_environ,
                  self.python_builtins, self.ayrton_builtins):
            try:
                ans= d[k]
                # found, don't search anymore (just in case you could find it
                # somewhere else)
                break
            except KeyError:
                strikes+= 1

        if strikes==5:
            # the name was not found in any of the dicts
            # create a command for it
            # ans= Command (k)
            # print (k)
            raise KeyError (k)

        return ans

    def __setitem__ (self, k, v):
        self.locals[k]= v

    def __delitem__ (self, k):
        del self.locals[k]

    def __iter__ (self):
        return self.locals.__iter__ ()

    def __str__ (self):
        return str ([ self.globals, self.locals, self.os_environ ])

class Ayrton (object):
    def __init__ (self, globals=None, locals=None, **kwargs):
        self.environ= Environment (globals, locals, **kwargs)
        self.options= {}

    def run_file (self, file):
        # it's a pity that parse() does not accept a file as input
        # so we could avoid reading the whole file
        self.run_script (open (file).read (), file)

    def run_script (self, script, file_name):
        tree= ast.parse (script)
        # ImportFrom(module='bar', names=[alias(name='baz', asname=None)], level=0)
        node= ImportFrom (module='ayrton.execute',
                          names=[alias (name='Command', asname=None)],
                          level=0)
        node.lineno= 0
        node.col_offset= 0
        ast.fix_missing_locations (node)
        tree.body.insert (0, node)
        tree= CrazyASTTransformer(self.environ).visit (tree)

        self.run_tree (tree, file_name)

    def run_tree (self, tree, file_name):
        self.run_code (compile (tree, file_name, 'exec'))

    def run_code (self, code):
        exec (code, self.environ.globals, self.environ)

def polute (d):
    # these functions will be loaded from each module and put in the globals
    # tuples (src, dst) renames function src to dst
    builtins= {
        'os': [ ('getcwd', 'pwd'), 'uname', 'listdir', ],
        'os.path': [ 'abspath', 'basename', 'commonprefix', 'dirname',  ],
        'time': [ 'sleep', ],
        'sys': [ 'exit' ], # argv is poluted from the CLI options. see Environment()

        'ayrton.file_test': [ '_a', '_b', '_c', '_d', '_e', '_f', '_g', '_h',
                              '_k', '_p', '_r', '_s', '_u', '_w', '_x', '_L',
                              '_N', '_S', '_nt', '_ot' ],
        'ayrton.expansion': [ 'bash', ],
        'ayrton.functions': [ 'cd', 'export', 'option', 'remote', 'run',
                               'shift', 'source', 'unset', ],
        'ayrton.execute': [ 'o', 'Capture', 'CommandFailed', 'CommandNotFound', ],
        }

    for module, functions in builtins.items ():
        m= importlib.import_module (module)
        for function in functions:
            if type (function)==tuple:
                src, dst= function
            else:
                src= function
                dst= function

            d[dst]= getattr (m, src)

    # now the IO files
    for std in ('stdin', 'stdout', 'stderr'):
        d[std]= getattr (sys, std).buffer

def run_tree (tree, globals, locals):
    global runner
    runner= Ayrton (globals=globals, locals=locals)
    runner.run_tree (tree)

def run_file_or_script (script=None, file=None, **kwargs):
    global runner
    runner= Ayrton (**kwargs)
    if script is None:
        runner.run_file (file)
    else:
        runner.run_script (script, 'script_from_command_line')

# backwards support for unit tests
main= run_file_or_script
